import unittest

from carball.analysis.analysis_manager import AnalysisManager

from ..json_parser.game import Game
from ..tests.utils import run_tests_on_list, run_analysis_test_on_replay

from .. import decompile_replays


class DBTest(unittest.TestCase):
    def test_replay_attrs(self):
        local = self

        def test(replay, file_path):
            json_object = decompile_replays.decompile_replay(replay, file_path)
            game = Game()
            game.initialize(loaded_json=json_object)
            info = game.game_info
            local.assertIsNotNone(game.game_info.server_name)
            local.assertIsNotNone(game.map)
            local.assertIsNotNone(game.game_info.match_guid)

        run_tests_on_list(test)

    def test_analysis(self):
        local = self

        def test(analysis: AnalysisManager):
            local.assertIsNotNone(analysis.get_protobuf_data())
            local.assertEqual(False, analysis.get_protobuf_data().game_metadata.is_invalid_analysis)

        run_analysis_test_on_replay(test)

if __name__ == '__main__':
    unittest.main()



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: api/stats/team_stats.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ...api.stats import stats_pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='api/stats/team_stats.proto',
  package='api.stats',
  serialized_pb=_b('\n\x1a\x61pi/stats/team_stats.proto\x12\tapi.stats\x1a\x15\x61pi/stats/stats.proto\"`\n\tTeamStats\x12)\n\npossession\x18\x01 \x01(\x0b\x32\x15.api.stats.Possession\x12(\n\nhit_counts\x18\x02 \x01(\x0b\x32\x14.api.stats.HitCounts')
  ,
  dependencies=[stats_pb2.DESCRIPTOR,])
_sym_db.RegisterFileDescriptor(DESCRIPTOR)




_TEAMSTATS = _descriptor.Descriptor(
  name='TeamStats',
  full_name='api.stats.TeamStats',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='possession', full_name='api.stats.TeamStats.possession', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='hit_counts', full_name='api.stats.TeamStats.hit_counts', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=64,
  serialized_end=160,
)

_TEAMSTATS.fields_by_name['possession'].message_type = stats_pb2._POSSESSION
_TEAMSTATS.fields_by_name['hit_counts'].message_type = stats_pb2._HITCOUNTS
DESCRIPTOR.message_types_by_name['TeamStats'] = _TEAMSTATS

TeamStats = _reflection.GeneratedProtocolMessageType('TeamStats', (_message.Message,), dict(
  DESCRIPTOR = _TEAMSTATS,
  __module__ = 'api.stats.team_stats_pb2'
  # @@protoc_insertion_point(class_scope:api.stats.TeamStats)
  ))
_sym_db.RegisterMessage(TeamStats)


# @@protoc_insertion_point(module_scope)



import os
from urllib import request
import json
from distutils.version import StrictVersion

response = request.urlopen('https://api.github.com/repos/tfausak/rattletrap/releases/latest')

js = json.loads(response.read())

cur_ver = '0.0.0'
binaries = [f for f in os.listdir('.') if not f.endswith('.py')]
print (binaries)
if len(binaries) > 0:
    cur_ver = binaries[0].split('-')[1]
update = StrictVersion(js['name']) > StrictVersion(cur_ver)
print (f'GitHub version: {js["name"]}\n'
       f'Current version: {cur_ver}\n'
       f'Update? {update}')
if update:
    for file in binaries:
        os.remove(file)

    for asset in js['assets']:

        print('Downloading {}'.format(asset['name']))
        request.urlretrieve(asset['browser_download_url'], asset['name'])


import json
import os
import subprocess
import logging
import platform as pt

from carball.analysis.analysis_manager import AnalysisManager
from carball.extras.per_goal_analysis import PerGoalAnalysis
from carball.json_parser.sanity_check.sanity_check import SanityChecker
from carball.json_parser.game import Game
from carball.controls.controls import ControlsCreator


logger = logging.getLogger(__name__)
BASE_DIR = os.path.dirname(__file__)


def decompile_replay(replay_path, output_path, overwrite=True):
    """
    Takes a path to the replay and outputs the json of that replay.

    :param replay_path: Path to a specific replay.
    :param output_path: The output path of rattletrap.
    :param overwrite: True if we should recreate the json even if it already exists.
    :return: The json created from rattle trap.
    """
    binaries = [f for f in os.listdir(os.path.join(BASE_DIR, 'rattletrap')) if not f.endswith('.py')]
    platform = pt.system()
    if platform == 'Windows':
        binary = [f for f in binaries if f.endswith('.exe')][0]
    elif platform == 'Linux':
        binary = [f for f in binaries if 'linux' in f][0]
    elif platform == 'Darwin':
        binary = [f for f in binaries if 'osx' in f][0]
    else:
        raise Exception('Unknown platform, unable to process replay file.')
    output_dirs = os.path.dirname(output_path)
    if not os.path.isdir(output_dirs) and output_dirs != '':
        os.makedirs(output_dirs)
    if overwrite or not os.path.isfile(output_path):
        cmd = [os.path.join(os.path.join(BASE_DIR, 'rattletrap'), '{}'.format(binary)), '--compact', '-i',
               replay_path,
               '--output',
               output_path]
        logger.debug(" ".join(cmd))
        subprocess.check_output(cmd)
    with open(output_path, encoding="utf8") as f:
        _json = json.load(f)
    return _json


def analyze_replay_file(replay_path: str, output_path: str, overwrite=True, controls: ControlsCreator=None,
                        sanity_check: SanityChecker=None, analysis_per_goal=False):
    """
    Decompile and analyze a replay file.

    :param replay_path: Path to replay file
    :param output_path: Path to write JSON
    :param overwrite: If to overwrite JSON (suggest True if speed is not an issue)
    :param controls: Generate controls from the replay using our best guesses (ALPHA)
    :param sanity_check: Run sanity check to make sure we analyzed correctly (BETA)
    :param analysis_per_goal: Runs the analysis per a goal instead of the replay as a whole
    :return: AnalysisManager of game with analysis.
    """
    _json = decompile_replay(replay_path, output_path, overwrite=overwrite)
    game = Game()
    game.initialize(loaded_json=_json)
    # get_controls(game)  # TODO: enable and optimise.
    if sanity_check is not None:
        sanity_check.check_game(game)
    if analysis_per_goal:
        analysis = PerGoalAnalysis(game)
    else:
        analysis = AnalysisManager(game)
    analysis.create_analysis()

    if controls is not None:
        controls.get_controls(game)

    return analysis


if __name__ == '__main__':
    from carball.tests.analysis_test import __test_replays
    __test_replays(BASE_DIR)



import os
import tempfile
from typing import Callable

import requests
from carball.decompile_replays import analyze_replay_file


def download_replay_discord(url):
    file = requests.get(url, stream=True)
    replay = file.raw.data
    return replay


def save_locally(replay_object):
    fd, file_path = tempfile.mkstemp()
    with open(file_path, 'wb') as f:
        f.write(replay_object)

    os.close(fd)
    return file_path


def run_tests_on_list(unit_test_func: Callable, replay_list=None, answers=None):
    if replay_list is None:
        replay_list = get_full_replay_list()

    for index in range(len(replay_list)):
        replay_url = replay_list[index]
        print('running test on replay: ' + replay_url[replay_url.rfind('/') + 1:])
        answer = answers[index] if answers is not None and index < len(answers) else None
        run_replay(replay_url, unit_test_func, answer=answer)


def run_replay(url, unit_test_func: Callable, answer=None):
    """
    Runs the replay with the file downloaded locally then deletes the file.
    :param url:
    :param unit_test_func:
    :param answer: data that can be passed to the replay to help judge it
    :return:
    """

    replay = download_replay_discord(url)
    file = save_locally(replay)

    fd, file_path = tempfile.mkstemp()
    os.close(fd)
    if answer is not None:
        unit_test_func(file, file_path, answer)
    else:
        unit_test_func(file, file_path)
    os.remove(file)
    os.remove(file_path)


def run_analysis_test_on_replay(unit_test_func: Callable, replay_list=None, answers=None):
    """
    :param unit_test_func: Called with an AnalysisManager
    :param replay_list: list of replay urls
    :return:
    """
    def wrapper(replay_file_path, json_file_path, answer=None):
        analysis_manager = analyze_replay_file(replay_file_path, json_file_path)
        if answer is not None:
            unit_test_func(analysis_manager, answer)
        else:
            unit_test_func(analysis_manager)
    run_tests_on_list(wrapper, replay_list, answers=answers)


def get_full_replay_list():
    """
    For full replays that have crashed or failed to be converted
    :return:
    """
    return [
        'https://cdn.discordapp.com/attachments/493849514680254468/493880540802449462/UnicodeEncodeError.replay',
        'https://cdn.discordapp.com/attachments/493849514680254468/496153554977816576/BOTS_JOINING_AND_LEAVING.replay',
        'https://cdn.discordapp.com/attachments/493849514680254468/496153569981104129/BOTS_NO_POSITION.replay',
        'https://cdn.discordapp.com/attachments/493849514680254468/496153605074845734/ZEROED_STATS.replay',
        'https://cdn.discordapp.com/attachments/493849514680254468/496180938968137749/FAKE_BOTS_SkyBot.replay'
    ]


def get_raw_replays():
    """
    Partial replays for specific values.
    :return:
    """
    return {
        "0_JUMPS": ["https://cdn.discordapp.com/attachments/493849514680254468/495735116895748096/0_JUMPS.replay"],
        "0_SAVES": ["https://cdn.discordapp.com/attachments/493849514680254468/495735137133264897/0_SAVES.replay"],
        "1_AERIAL": ["https://cdn.discordapp.com/attachments/493849514680254468/495735149133168651/1_AERIAL.replay"],
        "1_DEMO": ["https://cdn.discordapp.com/attachments/493849514680254468/495735163117109249/1_DEMO.replay"],
        "1_DOUBLE_JUMP": ["https://cdn.discordapp.com/attachments/493849514680254468/495735176761049103/1_DOUBLE_JUMP.replay"],
        "1_EPIC_SAVE": ["https://cdn.discordapp.com/attachments/493849514680254468/495735191537713153/1_EPIC_SAVE.replay"],
        "1_JUMP": ["https://cdn.discordapp.com/attachments/493849514680254468/495735203323576321/1_JUMP.replay"],
        "1_NORMAL_SAVE": ["https://cdn.discordapp.com/attachments/493849514680254468/495735215767945234/1_NORMAL_SAVE.replay"],
        # Boost
        "12_BOOST_PAD_0_USED": ["https://cdn.discordapp.com/attachments/493849514680254468/495735228422291456/12_BOOST_PAD_0_USED.replay"],
        "12_BOOST_PAD_45_USED": ["https://cdn.discordapp.com/attachments/493849514680254468/495735240321662986/12_BOOST_PAD_45_USED.replay"],
        "100_BOOST_PAD_0_USED": ["https://cdn.discordapp.com/attachments/493849514680254468/495735252233224193/100_BOOST_PAD_0_USED.replay"],
        "100_BOST_PAD_100_USED": ["https://cdn.discordapp.com/attachments/493849514680254468/495735264262488065/100_BOOST_PAD_100_USED.replay"],
        "NO_BOOST_PAD_0_USED": ["https://cdn.discordapp.com/attachments/493849514680254468/495735276338020372/NO_BOOST_PAD_0_USED.replay"],
        "NO_BOOST_PAD_33_USED": ["https://cdn.discordapp.com/attachments/493849514680254468/495735288254169109/NO_BOOST_PAD_33_USED.replay"],
        "12_AND_100_BOOST_PADS_0_USED": ["https://cdn.discordapp.com/attachments/493849514680254468/496071113768697859/12_AND_100_BOOST_PADS_0_USED.replay"],
        # Kickoffs
        "STRAIGHT_KICKOFF_GOAL": ["https://cdn.discordapp.com/attachments/493849514680254468/495735301604376576/Straight_Kickoff_Goal.replay"],
        "KICKOFF_NO_TOUCH": ["https://cdn.discordapp.com/attachments/493849514680254468/496034430943756289/NO_KICKOFF.replay"],
        "3_KICKOFFS": ["https://cdn.discordapp.com/attachments/493849514680254468/496034443442782208/3_KICKOFFS_4_SHOTS.replay"],
        # hits
        "4_SHOTS": ["https://cdn.discordapp.com/attachments/493849514680254468/496034443442782208/3_KICKOFFS_4_SHOTS.replay"],
        "KICKOFF_3_HITS": ["https://cdn.discordapp.com/attachments/493849514680254468/496072257928691743/KICKOFF_3_HITS.replay"],
        "MID_AIR_PASS": ["https://cdn.discordapp.com/attachments/493849514680254468/495887162928267314/MID_AIR_PASS_GOAL.replay"],
        "HIGH_AIR_PASS": ["https://cdn.discordapp.com/attachments/493849514680254468/495887164425633802/HIGH_AIR_PASS_GOAL.replay"],
        "GROUND_PASS": ["https://cdn.discordapp.com/attachments/493849514680254468/495887165570678794/GROUNDED_PASS_GOAL.replay"],
        "PINCH_GROUND": ["https://cdn.discordapp.com/attachments/493849514680254468/495887167932071947/PINCH_GROUNDED_GOAL.replay"],
    }


def get_specific_replays():
    raw_map = get_raw_replays()
    return {
        # BOOSTS
        "0_BOOST_COLLECTED": raw_map["NO_BOOST_PAD_0_USED"] + raw_map["NO_BOOST_PAD_33_USED"] + raw_map["KICKOFF_NO_TOUCH"],
        "1_SMALL_PAD": raw_map["12_BOOST_PAD_0_USED"] + raw_map["12_BOOST_PAD_45_USED"],
        "1_LARGE_PAD": raw_map["100_BOOST_PAD_0_USED"] + raw_map["100_BOST_PAD_100_USED"],
        "0_BOOST_USED": raw_map["12_BOOST_PAD_0_USED"] + raw_map["100_BOOST_PAD_0_USED"] + raw_map["NO_BOOST_PAD_0_USED"] + raw_map["KICKOFF_NO_TOUCH"],
        "BOOST_USED": raw_map["12_BOOST_PAD_45_USED"] + raw_map["100_BOST_PAD_100_USED"] + raw_map["NO_BOOST_PAD_33_USED"],  # [45, 100, 33]

        # HITS
        "HITS": raw_map["4_SHOTS"] + raw_map["KICKOFF_3_HITS"] + raw_map["12_BOOST_PAD_45_USED"] +
                raw_map["MID_AIR_PASS"] + raw_map["HIGH_AIR_PASS"] + raw_map["GROUND_PASS"] +
                raw_map["1_NORMAL_SAVE"] + raw_map["1_EPIC_SAVE"] + raw_map["1_AERIAL"],
               # + raw_map["PINCH_GROUND"],  TODO: Fix pinches to create 2 hits 1 for each person on same frame
        "SHOTS": raw_map["4_SHOTS"] + raw_map["12_BOOST_PAD_45_USED"] +
                 raw_map["1_EPIC_SAVE"] + raw_map["1_NORMAL_SAVE"],
        "PASSES": raw_map["MID_AIR_PASS"] + raw_map["HIGH_AIR_PASS"] + raw_map["GROUND_PASS"],
        "AERIALS": raw_map["1_EPIC_SAVE"] + raw_map["1_AERIAL"] + raw_map["HIGH_AIR_PASS"] + raw_map["MID_AIR_PASS"],
    }

def get_specific_answers():
    specific_replays = get_specific_replays()
    return {
        # Boost
        "0_BOOST_USED": [0] * len(specific_replays["0_BOOST_USED"]),
        "BOOST_USED": [45, 100, 33],
        # Hits
        "HITS": [4, 3, 1, 2, 9, 2, 4, 4, 4],
        "SHOTS": [3, 0, 2, 1],
        "PASSES": [1, 1, 1],
        "AERIALS": [0, 1, 2, 0]
    }




import json

import setuptools
from setuptools import setup
import os

with open(os.path.join('carball', 'analysis', 'PROTOBUF_VERSION'), 'r') as f:
    PROTOBUF_VERSION = json.loads(f.read())

subversion = 2
version_string = '0.' + str(PROTOBUF_VERSION) + '.' + str(subversion)

if os.path.isfile('README.md'):
    with open("README.md", "r") as readme_file:
        long_description = readme_file.read()
else:
    long_description = ''

setup(
    name='carball',
    version=version_string,
    packages=setuptools.find_packages(),
    include_package_data=True,
    install_requires=['pandas==0.23.4', 'protobuf==3.6.1', 'xlrd'],
    url='https://github.com/SaltieRL/ReplayAnalysis',
    keywords=['rocket-league'],
    license='Apache 2.0',
    author='Matthew Mage, Harry Xie, David Turner',
    author_email='sciguymjm@gmail.com',
    description='Rocket League replay parsing and analysis.',
    long_description=long_description,
    exclude_package_data={'': ['.gitignore', '.git/*', '.git/**/*', 'replays/*']}
)



from enum import Enum

import pandas as pd

from ..simulator.map_constants import MAP_Y


class FieldType(Enum):
    STANDARD = 1


HEIGHT_0_BALL_LIM = 95  # Height of car when on ground
HEIGHT_0_LIM = 20  # Height of car when on ground
HEIGHT_1_LIM = 840  # Goal height

MAP_THIRD = MAP_Y / 6

NEUTRAL_ZONE = MAP_Y / 20


class FieldConstants:

    field_type = FieldType.STANDARD

    def __init__(self, field_type=FieldType.STANDARD):
        self.field_type = field_type

    """
    All functions here assume the car is on the blue team and the game is played on a standard map.
    return: Boolean series that can be used to index the original data_frame to sum deltas with.
    """

    def get_neutral_zone(self, player_data_frame, **kwargs):
        return self.abs(player_data_frame.pos_y) < NEUTRAL_ZONE

    def get_half_0(self, player_data_frame, **kwargs):
        return player_data_frame.pos_y < 0

    def get_half_1(self, player_data_frame, **kwargs):
        return player_data_frame.pos_y > 0

    def get_third_0(self, player_data_frame, **kwargs):
        return player_data_frame.pos_y < -MAP_THIRD

    def get_third_1(self, player_data_frame, **kwargs):
        return (-MAP_THIRD < player_data_frame.pos_y) & (player_data_frame.pos_y < MAP_THIRD)

    def get_third_2(self, player_data_frame, **kwargs):
        return player_data_frame.pos_y > MAP_THIRD

    def get_height_0(self, player_data_frame, **kwargs):
        return player_data_frame.pos_z < HEIGHT_0_LIM

    def get_height_0_ball(self, player_data_frame, **kwargs):
        return player_data_frame.pos_z < HEIGHT_0_BALL_LIM

    def get_height_1(self, player_data_frame, **kwargs):
        return (HEIGHT_0_LIM < player_data_frame.pos_z) & (player_data_frame.pos_z < HEIGHT_1_LIM)

    def get_height_2(self, player_data_frame, **kwargs):
        return player_data_frame.pos_z > HEIGHT_1_LIM

    def get_ball_0(self, player_data_frame, ball_data_frame):
        """Ball is closer to goal 0 than player"""
        return player_data_frame.pos_y < ball_data_frame.pos_y

    def get_ball_1(self, player_data_frame, ball_data_frame):
        """Ball is closer to goal 1 than player"""
        return player_data_frame.pos_y > ball_data_frame.pos_y

    def abs(self, value):
        if value is pd.DataFrame:
            return value.abs()
        else:
            return abs(value)



import unittest

from carball.analysis.analysis_manager import AnalysisManager

from carball.tests.utils import run_analysis_test_on_replay, get_specific_replays, get_specific_answers


class BoostTest(unittest.TestCase):

    def test_1_small_pad_collected(self):

        def test(analysis: AnalysisManager):

            proto_game = analysis.get_protobuf_data()
            player = proto_game.players[0]
            boost = player.stats.boost
            self.assertEqual(boost.num_small_boosts, 1)
            print(analysis)

        run_analysis_test_on_replay(test, get_specific_replays()["1_SMALL_PAD"])

    def test_1_large_pad_collected(self):

        def test(analysis: AnalysisManager):

            proto_game = analysis.get_protobuf_data()
            player = proto_game.players[0]
            boost = player.stats.boost
            self.assertEqual(boost.num_large_boosts, 1)
            print(analysis)

        run_analysis_test_on_replay(test, get_specific_replays()["1_LARGE_PAD"])

    def test_0_boost_collected(self):

        def test(analysis: AnalysisManager):

            proto_game = analysis.get_protobuf_data()
            player = proto_game.players[0]
            boost = player.stats.boost
            self.assertEqual(boost.num_small_boosts, 0)
            self.assertEqual(boost.num_large_boosts, 0)
            print(analysis)

        run_analysis_test_on_replay(test, get_specific_replays()["0_BOOST_COLLECTED"])

    def test_boost_used(self):

        def test(analysis: AnalysisManager, boost_value):

            proto_game = analysis.get_protobuf_data()
            player = proto_game.players[0]
            boost = player.stats.boost
            self.assertAlmostEqual(boost.boost_usage, boost_value, delta=9) # TODO: Figgure out a way to calculate boost in a more accurate manner
            print(analysis)

        run_analysis_test_on_replay(test, get_specific_replays()["BOOST_USED"] + get_specific_replays()["0_BOOST_USED"],
                                    answers=get_specific_answers()["BOOST_USED"] +
                                            get_specific_answers()["0_BOOST_USED"])

    def test_0_used(self):

        def test(analysis: AnalysisManager, boost_value):

            proto_game = analysis.get_protobuf_data()
            player = proto_game.players[0]
            boost = player.stats.boost
            self.assertEqual(boost.boost_usage, boost_value)
            print(analysis)

        run_analysis_test_on_replay(test, get_specific_replays()["0_BOOST_USED"],
                                    answers=get_specific_answers()["0_BOOST_USED"])


if __name__ == '__main__':
    unittest.main()
